from collections import namedtuple
import json
from typing import List

from PyQt5 import QtCore
import jsonschema


DEF_LAYOUT_NAME = 'Default Layout Name'
DEF_MARGIN = 5.0
DEF_KEY_WIDTH = 30.0
DEF_KEY_HEIGHT = 35.0
DEF_FONT = ''
DEF_FONT_COLOR = '#000000'
DEF_BACKGROUND_COLOR = '#FFFFFF'
DEF_STROKE_COLOR = '#000000'
DEF_KEY_COLOR = ''
DEF_KEY_COLOR_PRESSED = '#000000'

DEF_KEY_LABEL = ''
DEF_KEY_WIDTH_UNIT = 1.0
DEF_KEY_HEIGHT_UNIT = 1.0
DEF_KEY_POS_X = 0.0
DEF_KEY_POS_Y = 0.0

StenoKey = namedtuple('StenoKey',
                      'name label ' +
                      'x y ' +
                      'width height ' +
                      'is_round_top is_round_bottom ' +
                      'font_color stroke_color color color_pressed')

class StenoLayout():
    ''' Represents the structure of a stenography layout '''

    def __init__(self):
        self.name = DEF_LAYOUT_NAME
        self.margin = DEF_MARGIN
        self.key_width = DEF_KEY_WIDTH
        self.key_height = DEF_KEY_HEIGHT
        self.font = DEF_FONT
        self.background_color = DEF_BACKGROUND_COLOR

        self.keys: List[StenoKey] = []

        self._validation_schema = StenoLayout._load_validation_schema()

    def load_from_file(self, file_path: str) -> bool:
        ''' Populates layout information from the provided file path '''

        try:
            with open(file_path, encoding='utf-8') as file:
                data = json.load(file)
        except:
            return False

        return self.load_from_json(data)

    def load_from_resource(self, resource_path: str) -> bool:
        ''' Populates layout information from the provided Qt resource path '''

        try:
            file_text = StenoLayout._load_qt_resource_text(resource_path)
            data = json.loads(file_text)
        except:
            return False

        return self.load_from_json(data)

    def load_from_json(self, data: dict) -> bool:
        ''' Populates layout information from the provided JSON data '''

        # Validate the JSON file first before trying to load it
        try:
            jsonschema.validate(data, self._validation_schema)
        except (jsonschema.ValidationError, jsonschema.SchemaError) as error:
            return False

        self.name = data.get('name', DEF_LAYOUT_NAME)
        self.margin = data.get('margin', DEF_MARGIN)
        self.key_width = data.get('key_width', DEF_KEY_WIDTH)
        self.key_height = data.get('key_height', DEF_KEY_HEIGHT)
        self.font = data.get('font', DEF_FONT)
        font_color = data.get('font_color', DEF_FONT_COLOR)
        self.background_color = data.get('background_color', DEF_BACKGROUND_COLOR)
        key_stroke_color = data.get('key_stroke_color', DEF_STROKE_COLOR)
        key_color = data.get('key_color', DEF_KEY_COLOR)
        key_color_pressed = data.get('key_color', DEF_KEY_COLOR_PRESSED)

        self.keys = []
        for key in data['keys']:
            self.keys.append(StenoKey(
                # Make sure that name always exists; we'll default to unique
                key.get('name', str(len(self.keys))),
                key.get('label', DEF_KEY_LABEL),
                key.get('x', DEF_KEY_POS_X),
                key.get('y', DEF_KEY_POS_Y),
                key.get('width', DEF_KEY_WIDTH_UNIT),
                key.get('height', DEF_KEY_HEIGHT_UNIT),
                key.get('is_round_top', False),
                key.get('is_round_bottom', False),
                key.get('font_color', font_color),
                key.get('stroke_color', key_stroke_color),
                key.get('color', key_color),
                key.get('color_pressed', key_color_pressed)
            ))

        return True

    @staticmethod
    def _load_validation_schema() -> dict:
        ''' Loads a JSON Schema validation schema '''

        data = {}

        try:
            file_text = StenoLayout._load_qt_resource_text(':/layout_display/steno_layout.schema.json')
            data = json.loads(file_text)
        except:
            return {}

        return data

    @staticmethod
    def _load_qt_resource_text(resource_path: str) -> str:
        ''' Wrapper for loading a Qt resource file's text contents '''

        file_text = ''

        try:
            file = QtCore.QFile(resource_path)

            if file.open(QtCore.QIODevice.ReadOnly | QtCore.QFile.Text):
                text_stream = QtCore.QTextStream(file)
                text_stream.setCodec('utf-8')
                text_stream.setAutoDetectUnicode(True)

                file_text = str(text_stream.readAll())
                file.close()
        except:
            return ''

        return file_text



from pathlib import Path
from typing import Optional

from PyQt5.QtCore import QSettings
from PyQt5.QtWidgets import QFileDialog

from plover import system
from plover.oslayer.config import CONFIG_DIR
from plover.engine import StenoEngine
from plover.steno import Stroke
from plover.gui_qt.tool import Tool
from plover.gui_qt.utils import ToolBar

from layout_display.layout_display_ui import Ui_LayoutDisplay
from layout_display.steno_layout import StenoLayout


class LayoutDisplay(Tool, Ui_LayoutDisplay):
    ''' Stenography layout display of strokes. '''

    TITLE = 'Layout Display'
    ROLE = 'layout_display'
    ICON = ':/layout_display/steno_key.svg'

    def __init__(self, engine: StenoEngine):
        super(LayoutDisplay, self).__init__(engine)
        self.setupUi(self)

        self.layout().addWidget(ToolBar(
            self.action_Reset,
            self.action_Load
        ))

        self._stroke = []
        self._numbers = set()
        self._numbers_to_keys = {}
        self._number_key = ''
        self._system_name = ''
        self._system_file_map = {}

        self._layout = StenoLayout()

        self.restore_state()
        self.finished.connect(self.save_state)

        engine.signal_connect('config_changed', self.on_config_changed)
        self.on_config_changed(engine.config)
        engine.signal_connect('stroked', self.on_stroke)

        self.layout_display_view.initialize_view(self._layout)

    def _save_state(self, settings: QSettings):
        '''
        Save state to settings.
        Called via save_state through plover.qui_qt.utils.WindowState
        '''

        settings.setValue('system_file_map', self._system_file_map)

    def _restore_state(self, settings: QSettings):
        '''
        Restore state from settings.
        Called via restore_state through plover.qui_qt.utils.WindowState
        '''

        self._system_file_map = settings.value('system_file_map', {}, dict)

    def on_config_changed(self, config):
        ''' Updates state based off of the new Plover configuration '''

        # If something unrelated changes like a new dictionary
        # being added then the system name will not be in config
        if 'system_name' not in config:
            return

        self._stroke = []
        self._numbers = set(system.NUMBERS.values())
        self._numbers_to_keys = {v: k for k, v in system.NUMBERS.items()}
        self._number_key = system.NUMBER_KEY
        self._system_name = config['system_name']

        # If the user has no valid preference then fall back to the default
        preferred_layout_file = self.get_preferred_layout(self._system_name)
        if preferred_layout_file and self._layout.load_from_file(preferred_layout_file):
            self.label_layout_name.setText(self._layout.name)
            self.layout_display_view.initialize_view(self._layout)
        else:
            self.on_reset()

    def on_stroke(self, stroke: Stroke):
        ''' Updates state based off of the latest stroke by the user '''

        keys = stroke.steno_keys[:]

        # Handle converting numbers in the stroke to the actual key values
        if any(key in self._numbers for key in keys):
            keys.append(self._number_key)
        keys = [self._numbers_to_keys[x] if x in self._numbers_to_keys else x for x in keys]

        self._stroke = keys
        self.layout_display_view.update_view(self._layout, keys)

    def on_reset(self):
        ''' Resets the layout to the built-in default layout '''

        self._remove_system_file_map(self._system_name)

        self._layout.load_from_resource(':/layout_display/english_stenotype.json')
        self.label_layout_name.setText(self._layout.name)
        self.layout_display_view.initialize_view(self._layout)

    def on_load(self):
        ''' Gets a layout file from the user to load '''

        # The API says this should return a string, but it returns a tuple

        file_path, _filter = QFileDialog.getOpenFileName(self, 'Open Layout File',
                                                         CONFIG_DIR, '(*.json)')

        # If the user cancelled out of the dialog then we will have a null string
        if not file_path:
            return

        if self._layout.load_from_file(file_path):
            self._add_system_file_map(self._system_name, file_path)
            self.label_layout_name.setText(self._layout.name)
            self.layout_display_view.initialize_view(self._layout)
        else:
            self.on_reset()

    def get_preferred_layout(self, system_name: str) -> Optional[str]:
        ''' Gets the user's preferred layout file for the given system '''

        file_path = self._system_file_map.get(system_name)

        # At least validate the file still exists
        if file_path and not Path(file_path).is_file():
            file_path = None
            self._remove_system_file_map(system_name)
            self.save_state()

        return file_path

    def _add_system_file_map(self, system_name: str, file_path: str):
        ''' Updates the mapping from system to file path with a new entry '''

        self._system_file_map[system_name] = file_path
        self.save_state()

    def _remove_system_file_map(self, system_name: str):
        ''' Updates the mapping from system to file path by removing an entry '''

        if system_name in self._system_file_map:
            self._system_file_map.pop(system_name)
            self.save_state()



from typing import List

from PyQt5.QtCore import Qt, QMarginsF
from PyQt5.QtWidgets import (QWidget, QGraphicsView, QGraphicsScene,
                             QGraphicsTextItem)
from PyQt5.QtGui import (QPainterPath, QPen, QBrush, QColor,
                         QFont, QResizeEvent, QShowEvent)

from layout_display.steno_layout import StenoLayout, StenoKey


class LayoutDisplayView(QGraphicsView):
    ''' View to manage Layout Display scenes '''

    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self.graphics_scene = LayoutDisplayScene(self)
        self._scene_pen = QPen(QColor('#000000'), 1, Qt.SolidLine,
                               Qt.SquareCap, Qt.BevelJoin)

    def resizeEvent(self, event: QResizeEvent):
        ''' Qt resize event '''

        # Make sure the layout display scene scales to the view
        self.fitInView(self.graphics_scene.sceneRect(), Qt.KeepAspectRatio)

    def showEvent(self, event: QShowEvent):
        ''' Qt show event '''

        # Make sure the layout display scene is scaled on initial "show"
        self.fitInView(self.graphics_scene.sceneRect(), Qt.KeepAspectRatio)

    def initialize_view(self, steno_layout: StenoLayout):
        ''' Initializes the view for the provided layout '''

        # TODO: For now, just call update without a stroke, but these two
        #       should be optimized to not re-create the scene each time
        self.update_view(steno_layout, [])

    def update_view(self, steno_layout: StenoLayout, stroke: List[str]):
        ''' Updates the layout display for the provided layout and stroke '''

        scene = self.graphics_scene
        pen = self._scene_pen
        font = QFont(steno_layout.font) if steno_layout.font else QFont()
        # Clear all items from the scene. Could be more efficient...
        scene.clear()
        scene.setBackgroundBrush(QBrush(QColor(steno_layout.background_color)))

        for key in steno_layout.keys:
            path = LayoutDisplayView._create_key_path(steno_layout, key)
            brush = LayoutDisplayView._get_key_path_brush(key, (key.name in stroke))
            pen.setColor(QColor(key.stroke_color))

            # Add the key path before its label, then center the label
            scene.addPath(path, pen, brush)

            if key.label:
                label = QGraphicsTextItem(key.label)
                label.setFont(font)
                label.setDefaultTextColor(QColor(key.font_color))

                label_rect = label.boundingRect()
                label_rect.moveCenter(path.boundingRect().center())
                label.setPos(label_rect.x(), label_rect.y())
                scene.addItem(label)

        # Scene rects don't shrink when items are removed, so need to manually
        # set it to the current size needed by the contained items + margin
        margin = steno_layout.margin
        scene_rect = scene.itemsBoundingRect()
        scene_rect = scene_rect.marginsAdded(QMarginsF(margin, margin,
                                                       margin, margin))
        scene.setSceneRect(scene_rect)
        self.fitInView(scene.sceneRect(), Qt.KeepAspectRatio)

        self.setScene(scene)
        self.show()

    @staticmethod
    def _get_key_path_brush(key: StenoKey, is_pressed: bool) -> QBrush:
        ''' Gets what brush the key should use '''

        if is_pressed and key.color_pressed:
            path_brush = QBrush(QColor(key.color_pressed))
        elif not is_pressed and key.color:
            path_brush = QBrush(QColor(key.color))
        else:
            path_brush = QBrush()

        return path_brush

    @staticmethod
    def _create_key_path(steno_layout: StenoLayout, key: StenoKey) -> QPainterPath:
        ''' Creates the path for a key '''

        key_width = steno_layout.key_width
        key_height = steno_layout.key_height

        pos_x = key.x * key_width
        pos_y = key.y * key_height
        width = key_width * key.width
        height = key_height * key.height
        is_round_top = key.is_round_top
        is_round_bottom = key.is_round_bottom

        # Figure out adjustments needed to add rounded parts
        ellipse_height = min((width, height / 2))

        pos_y = pos_y + ellipse_height / 2 if is_round_top else pos_y
        if is_round_top and is_round_bottom:
            height = height - ellipse_height
        elif is_round_top or is_round_bottom:
            height = height - ellipse_height / 2

        # Construct the main key shape
        key_path = QPainterPath()
        key_path.addRect(pos_x, pos_y, width, height)

        # Add on the rounded parts
        if is_round_top:
            key_top = QPainterPath()

            key_top.addEllipse(pos_x, pos_y - ellipse_height / 2,
                               width, ellipse_height)
            key_path = key_path.united(key_top)
        if is_round_bottom:
            key_bottom = QPainterPath()

            key_bottom.addEllipse(pos_x, pos_y + height - ellipse_height / 2,
                                  width, ellipse_height)
            key_path = key_path.united(key_bottom)

        return key_path

class LayoutDisplayScene(QGraphicsScene):
    ''' Scene containing all layout display objects '''

    def __init__(self, parent: QWidget):
        super().__init__(parent)
